/************************************************************
 **
 ** pic_var.c : PIC variable allocation definitions
 **
 ** Copyright (c) 2004-2005, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>

#include "../libutils/mem.h"
#include "../libcore/pf_proc.h"
#include "pic_msg.h"
#include "pic_stvar.h"
#include "pic_var.h"

#undef BLIST_DEBUG

static unsigned long pic_var_alloc_pass;

unsigned         pic_blist_max; /* max amount used     */
unsigned         pic_blist_cur; /* current amount used */
unsigned         pic_blist_total;
pic_bank_info_t *pic_blist;

size_t           pic_stk_sz;
variable_base_t  pic_stk_base;

#ifdef BLIST_DEBUG
static void pic_blist_dump(pfile_t *pf, pic_bank_info_t *blist)
{
  const pic_bank_info_t *bptr;

  pfile_log(pf, pfile_log_debug, "--- blist start ---");
  for (bptr = blist; bptr; bptr = bptr->link) {
    pfile_log(pf, pfile_log_debug, 
        PIC_MSG_DBG_BANK, bptr->lo, bptr->hi);
  }
  pfile_log(pf, pfile_log_debug, "--- blist end ---");
}
#endif


/*
 * NAME
 *   pic_variable_alloc
 *
 * DESCRIPTION
 *   allocate space for all variables
 *
 * PARAMETERS
 *   pf
 *
 * RETURN
 *
 * NOTES
 *   this is dummied up for now. eventually i'll need to sort
 *   the variables by size for a more optimimal filling
 */
/* this is the union of  blist with {lo,hi} */
void pic_blist_add(pfile_t *pf, pic_bank_info_t **pblist, 
    size_t lo, size_t hi)
{
  pic_bank_info_t *ptr;
  pic_bank_info_t *pv;
  pic_bank_info_t *blist;

  assert(lo <= hi);
  blist = *pblist;
  for (ptr = blist, pv = 0; 
       ptr && (lo > ptr->hi);
       pv = ptr, ptr = ptr->link)
    ;
  /* this could use *a lot* of improvement, but will work for now
   * 1st I'll simply allocate a new entry & insert it, then I'll
   * do some cleanup */
  if (ptr && (ptr->lo <= lo) && (ptr->hi >= hi)) {
  } else {
    ptr = MALLOC(sizeof(*ptr));
    ptr->lo = lo;
    ptr->hi = hi;
    ptr->link = (pv) ? pv->link : blist;
    if (pv) {
      ptr->link = pv->link;
      pv->link  = ptr;
    } else {
      ptr->link = blist;
      blist     = ptr;
    }

    ptr = blist;
    while (ptr->link) {
      pic_bank_info_t *link;

      link = ptr->link;
      if (ptr->hi + 1 >= link->lo) {
        /* remove ptr->link */
        if (ptr->hi < link->hi) {
          ptr->hi = link->hi;
        }
        ptr->link = link->link;
        FREE(link);
      } else {
        ptr = link;
      }
    }
  }
  *pblist = blist;
#ifdef BLIST_DEBUG
  pfile_log(pf, pfile_log_debug, "BLIST ADD(%u,%u)", lo, hi);
  pic_blist_dump(pf, pic_blist);
#endif
}

static void pic_blist_remove(pfile_t *pf, pic_bank_info_t **pblist,
    size_t base, size_t sz)
{
  pic_bank_info_t *ptr;
  pic_bank_info_t *pv;
  pic_bank_info_t *blist;
  size_t           lo;
  size_t           hi;

  blist = *pblist;
  lo = base;
  hi = base + sz - 1;

  for (ptr = blist, pv = 0;
       ptr && (lo > ptr->hi);
       pv = ptr, ptr = ptr->link)
    ;
  if (ptr && (lo <= ptr->hi) && (hi >= ptr->lo)) {
    /* lo <= ptr->hi
       ptr->lo <= hi <= ptr->hi */
    if (lo <= ptr->lo) {
      ptr->lo = hi + 1;
    }
    if (hi >= ptr->hi) {
      ptr->hi = lo - 1;
    }
    /* if (ptr->hi < ptr->lo), this bit is gone */
    if (ptr->hi < ptr->lo) {
      if (pv) {
        pv->link = ptr->link;
      } else {
        *pblist = ptr->link;
      }
      FREE(ptr);
    } else if ((ptr->lo < lo) && (hi < ptr->hi)) {
      /* split this into:
           {ptr->lo, lo - 1}
           {hi + 1, ptr->hi} */
      pv = ptr;
      ptr = MALLOC(sizeof(*ptr));
      if (!ptr) {
        pfile_log_syserr(pf, result_memory);
      } else {
        ptr->link = pv->link;
        pv->link  = ptr;
        ptr->hi   = pv->hi;
        ptr->lo   = hi + 1;
        pv->hi    = lo - 1;
      }
    }
  }
  *pblist = blist;
#ifdef BLIST_DEBUG
  pfile_log(pf, pfile_log_debug, "BLIST REMOVE(%u,%u)", lo, lo + sz - 1);
  pic_blist_dump(pf, pic_blist);
#endif
}

/* allocate a block out of the data area */
static variable_base_t pic_data_alloc(pfile_t *pf, pic_bank_info_t **blist,
  variable_sz_t need)
{
  pic_bank_info_t *bank;
  variable_base_t  base;

  base = VARIABLE_BASE_UNKNOWN;

  for (bank = *blist; 
       bank && (bank->lo + need > bank->hi); 
       bank = bank->link)
    ;
  if (bank) {
    base = bank->lo;
    pic_blist_remove(pf, blist, bank->lo, need);
#ifdef BLIST_DEBUG
    fprintf(stderr, "Allocating %u\n", base);
#endif
  } 
  return base;
}

/* returns boolean_true if a new base was set, else boolean_false */
static boolean_t pic_variable_base_set(pfile_t *pf, pic_bank_info_t **blist,
    variable_t vptr)
{
  size_t          need;
  variable_base_t base;
  boolean_t       rc;

  need = variable_sz_get(vptr);
  assert(need);
  /* first we'll see if the entire thing can fit in a bank.
     If so, good! */

  base = pic_data_alloc(pf, blist, need);
  rc   = boolean_false;
  if (VARIABLE_BASE_UNKNOWN == base) {
    pfile_log(pf, pfile_log_err, "no space for %s%u",
        variable_name_get(vptr), variable_tag_n_get(vptr));
  } else {
    if ((VARIABLE_BASE_UNKNOWN == variable_base_get(vptr, 0))
      || (base > variable_base_get(vptr, 0))) {
      rc = boolean_true;
      variable_base_set(vptr, base, 0);
      variable_pass_set(vptr, pic_var_alloc_pass);
#if 0 /* def BLIST_DEBUG */
      fprintf(stderr, "Setting %x:%s:%u to 0x%x in pass %u\n",
          (unsigned) vptr, variable_name_get(vptr),
          variable_tag_n_get(vptr),
          base, pic_var_alloc_pass);
#endif
    } else {
      size_t lo;
      size_t hi;

      lo = base;
      hi = lo + need - 1;
      pic_blist_add(pf, &pic_blist, lo, hi);
    }
  }
  return rc;
}

void pic_variable_alloc_one(pfile_t *pf, variable_t var)
{
  pic_variable_base_set(pf, &pic_blist, var);
}

/* find all variables that the user has placed & make sure these
   positions do not get reused */
static void blist_total(void)
{
  const pic_bank_info_t *binfo;

  for (pic_blist_total = 0, binfo = pic_blist; 
       binfo; 
       pic_blist_total += binfo->hi - binfo->lo + 1, binfo = binfo->link)
    ;
}

static void pic_variable_alloc_proc(pfile_t *pf, pfile_proc_t *proc);
static void pic_variable_alloc_block(pfile_t *pf, pfile_proc_t *proc, 
    pfile_block_t *blk)
{
  variable_t     vptr;
  pfile_block_t *cblk;
  size_t         ct;
  pfile_proc_t  *cproc;
  unsigned long  pass;
  size_t         entry_sz;

  entry_sz = pic_blist_cur;

  pass = ++pic_var_alloc_pass;

  if (!blk) {
    blk = pfile_proc_block_root_get(proc);
  }
  /* first, allocate all variables for this procedure */
  for (vptr = pfile_block_variable_list_head(blk); 
       vptr; 
       vptr = variable_link_get(vptr)) {
    if (variable_is_auto(vptr)) {
      /* find the first bank that will hold this variable */
      size_t sz;
      variable_base_t old_base;

#if 0
      fprintf(stdout, "Allocating...%s(%s%u)\n",
          pfile_proc_tag_get(proc), variable_name_get(vptr),
          variable_tag_n_get(vptr));
#endif
      sz = variable_sz_get(vptr);
      assert(sz);
      old_base = variable_base_get(vptr, 0);
      if (pic_variable_base_set(pf, &pic_blist, vptr)) {
        if (old_base != VARIABLE_BASE_UNKNOWN) {
          pfile_log(pf, pfile_log_debug, 
              "Resetting...%s%u (was %u, now %u)",
              variable_name_get(vptr), variable_tag_n_get(vptr),
              old_base, variable_base_get(vptr, 0)
              );
        }
        pic_blist_cur += sz;
        if (pic_blist_cur > pic_blist_max) {
          pic_blist_max = pic_blist_cur;
        }
#if 0
      } else {
        label_t lbl;

        lbl = pfile_proc_label_get(proc);
        pfile_log(pf, pfile_log_err, "cannot place %s:%s", 
            (lbl) ? label_name_get(lbl) : "{global}", 
            variable_name_get(vptr));
#endif
      }
    }
  }
  /* next, recurse into all procedures this one calls and
   * allocate *those* variables */
  for (cblk = pfile_block_child_get(blk);
       cblk;
       cblk = pfile_block_sibbling_get(cblk)) {
    pic_variable_alloc_block(pf, proc, cblk);
  }
  for (ct = 0; (cproc = pfile_proc_calls_get(proc, ct)) != 0; ct++) {
    if (!pfile_proc_flag_test(cproc, PFILE_PROC_FLAG_VISITED)) {
      pfile_proc_flag_set(cproc, PFILE_PROC_FLAG_VISITED);
      pic_variable_alloc_proc(pf, cproc);
      pfile_proc_flag_clr(cproc, PFILE_PROC_FLAG_VISITED);
    }
  }
  /* determine the largest available block & size for the stack pointer */
  if (pic_blist_max == pic_blist_cur) {
    const pic_bank_info_t *pic_blist_ptr;

    for (pic_blist_ptr = pic_blist;
         pic_blist_ptr;
         pic_blist_ptr = pic_blist_ptr->link) {
      size_t sz;

      sz = pic_blist_ptr->hi + 1 - pic_blist_ptr->lo;
      if (sz > pic_stk_sz) {
        pic_stk_base = pic_blist_ptr->hi + 1;
        pic_stk_sz   = sz;
      }
    }
  }
  /* finally, return any AUTO variables back to the blist */
  for (vptr = pfile_block_variable_list_head(blk);
       vptr;
       vptr = variable_link_get(vptr)) {
    if (variable_is_auto(vptr)
        && (VARIABLE_BASE_UNKNOWN != variable_base_get(vptr, 0))
        && (variable_pass_get(vptr) == pass)) {
      size_t lo;
      size_t hi;
      size_t sz;
#if 0
      fprintf(stdout, "Releasing %x:%s:%u from %u in pass %u\n",
          (unsigned) vptr, variable_name_get(vptr),
          variable_tag_n_get(vptr),
          variable_base_get(vptr), pass);
#endif
      lo = variable_base_get(vptr, 0);
      sz = variable_sz_get(vptr);
      hi = lo + sz - 1;
      if (sz > pic_blist_cur) {
        fprintf(stderr, "invalid variable = %x:%s %u\n", 
            (unsigned) vptr, variable_name_get(vptr), variable_tag_n_get(vptr));
      } 
      assert(sz <= pic_blist_cur);
      pic_blist_cur -= sz;
      pic_blist_add(pf, &pic_blist, lo, hi);
      variable_pass_set(vptr, -1);
#ifdef BLIST_DEBUG
      fprintf(stderr, "Returning %s\n", variable_name_get(vptr));
#endif
    }
  }
  {
    static unsigned ctr;

    ctr++;
    assert(pic_blist_cur == entry_sz);
  }
}

static void pic_variable_alloc_proc(pfile_t *pf, pfile_proc_t *proc)
{
  /*
   * if the procedure is ever called indirectly its variables were
   * allocated earlier so this can be ignored
   */
  if (!pfile_proc_flag_test(proc, PFILE_PROC_FLAG_INDIRECT)) {
    pic_variable_alloc_block(pf, proc, 0);
  }
}

/* allocate all variables out of a single block. base holds the block */
static void pic_block_data_alloc(pfile_block_t *blk, variable_base_t base)
{
  variable_t vptr;

  for (vptr = pfile_block_variable_list_head(blk);
       vptr;
       vptr = variable_link_get(vptr)) {
    if (variable_is_auto(vptr)) {
      variable_base_set(vptr, base, 0);
      base += variable_sz_get(vptr);
    }
  }
  for (blk = pfile_block_child_get(blk);
       blk;
       blk = pfile_block_sibbling_get(blk)) {
    pic_block_data_alloc(blk, base);
  }
}

/* all interrupt entry points share the same variable space, so...
     1. calculate the maximum size needed
     2. allocate a single block that size
     3. divy it up appropriately
 */
static void pic_variable_alloc_isr(pfile_t *pf)
{
  variable_base_t isr_block;
  pfile_proc_t   *proc;
  size_t          sz;

  for (proc = pfile_proc_root_get(pf), sz = 0; 
       proc; 
       proc = pfile_proc_next(proc)) {
    if (pfile_proc_flag_test(proc, PFILE_PROC_FLAG_INTERRUPT) 
      && (sz < pfile_proc_frame_sz_get(proc))) {
      sz = pfile_proc_frame_sz_get(proc);
    }
  }

  isr_block = (sz)
    ? pic_data_alloc(pf, &pic_blist, sz)
    : VARIABLE_BASE_UNKNOWN;

  for (proc = pfile_proc_root_get(pf);
       proc;
       proc = pfile_proc_next(proc)) {
    if (pfile_proc_flag_test(proc, PFILE_PROC_FLAG_INTERRUPT)) {
      pfile_block_t *blk;

      /* allocate all of the procedure's variables out of isr_block */
      pic_block_data_alloc(pfile_proc_block_root_get(proc), isr_block);
      /* remove the AUTO flag from all of the procedure's variables */
      for (blk = pfile_proc_block_root_get(proc);
           blk;
           blk = pfile_block_next(blk)) {
        variable_t vptr;

        for (vptr = pfile_block_variable_list_head(blk);
             vptr;
             vptr = variable_link_get(vptr)) {
          variable_flag_clr(vptr, VARIABLE_FLAG_AUTO);
        }
      }
      /* allocate anything called by this proc */
      pic_variable_alloc_proc(pf, proc);
    }
  }
}

/*
 * allocate variables for all of the indirect functions. It's probably
 * to analyze these a bit better & do a smarter allocation so we can
 * reuse data space, but it's far easier to just say dataspace for all 
 * functions called indirectly cannot be reused
 */
static void pic_variable_alloc_indirect(pfile_t *pf)
{
  pfile_proc_t *proc;

  for (proc = pfile_proc_root_get(pf);
       proc;
       proc = pfile_proc_next(proc)) {
    if (pfile_proc_flag_test(proc, PFILE_PROC_FLAG_INDIRECT)
#if 0
        && (pfile_proc_flag_test(proc, PFILE_PROC_FLAG_CONTEXT_USER)
          || pfile_proc_flag_test(proc, PFILE_PROC_FLAG_CONTEXT_ISR))) 
#endif
        && pfile_proc_frame_sz_get(proc)) {
      variable_base_t base;

      base = pic_data_alloc(pf, &pic_blist, pfile_proc_frame_sz_get(proc));
      pic_block_data_alloc(pfile_proc_block_root_get(proc), base);
    }
  }
}

void pic_variable_alloc(pfile_t *pf)
{
  pfile_proc_t *proc;

  blist_total();
  pfile_log(pf, pfile_log_debug, PIC_MSG_ALLOCATING_VARS);
  pic_stvar_fixup(pf);

  /* make sure the frame sizes are correct. currently only re-entrant
   * functions create an extra variable but that may change */

  for (proc = pfile_proc_root_get(pf);
       proc;
       proc = pfile_proc_next(proc)) {
    pfile_proc_frame_sz_calc(proc);
  }
  /* remove any preset variables */
  for (proc = pfile_proc_root_get(pf);
       proc;
       proc = pfile_proc_next(proc)) {
    pfile_block_t *blk;

    for (blk = pfile_proc_block_root_get(proc);
         blk;
         blk = pfile_block_next(blk)) {
      variable_t    vptr;

      for (vptr = pfile_block_variable_list_head(blk);
           vptr; 
           vptr = variable_link_get(vptr)) {
        if (!variable_master_get(vptr)
            && (VARIABLE_BASE_UNKNOWN != variable_base_get(vptr, 0))) {
          variable_base_t base;

          if (variable_sz_get(vptr)) {
            size_t ii;

            for (ii = 0; ii < VARIABLE_MIRROR_CT; ii++) {
              base = variable_base_get(vptr, 0);
              if (VARIABLE_BASE_UNKNOWN != base) {
                pic_blist_remove(pf, &pic_blist, variable_base_get(vptr, 0),
                      variable_sz_get(vptr));
              }
            }
          }
        }
      }
    }
  }
  pic_variable_alloc_isr(pf);
  pic_variable_alloc_indirect(pf);
  /* finally, allocate some variables! */
  pic_variable_alloc_proc(pf, pfile_proc_root_get(pf));
  pfile_log(pf, pfile_log_debug, PIC_MSG_DBG_DATA_USED, pic_blist_max);
  if (pic_blist_cur) {
    fprintf(stderr, "pic_blist_cur = %u\n", pic_blist_cur);
  }
  assert(0 == pic_blist_cur);
}

void pic_blist_info_log(pfile_t *pf)
{
#if 0
  pic_bank_info_t *bank;
  unsigned         total_avail;
  unsigned         total_used;
  unsigned         ct;

  for (ct = 0, total_used = 0, total_avail = 0, bank = pic_blist;
       bank;
       ct++, total_used += bank->next - bank->lo, 
         total_avail += bank->hi + 1 - bank->lo, bank = bank->link)
    ;
  pfile_log(pf, pfile_log_info, PIC_MSG_DATA_USED,
      total_used, total_avail);
#endif
}

void pic_blist_free(pfile_t *pf)
{
  while (pic_blist) {
    pic_bank_info_t *bank;

    bank = pic_blist->link;
    FREE(pic_blist);
    pic_blist = bank;
  }
}

pic_bank_info_t *pic_bank_info_get(pfile_t *pf)
{
  return pic_blist;
}

